@@ -1,5 +1,8 @@
 """
     通过 redis 实现心跳功能
+
+    master 异常会把自己设置为 slave
+    slave 检测到 master 死机会把自己设置为 master
 """
 import json
 import time
@@ -19,6 +22,7 @@ class ClientHeart:
         """
         self.spider = spider
 
+        self.stop = False  # 停止标志（用于异常时）
         self.beating_time = 4  # 心跳频率
         self.check_time = self.beating_time - 1  # 心跳检查频率
         self.client_name = self.generate_client_name()  # 随机客户端名
@@ -29,6 +33,10 @@ class ClientHeart:
 
         try:
             self.check_client_beating()
+        except Exception as e:
+            self.stop = True
+            self.spider.spider_master = False
+            logger.exception(e)
         finally:
             self.close(beating)
 
@@ -82,7 +90,7 @@ class ClientHeart:
                         logger.debug(f"心跳正常：{client_name}")
 
                 # 如果所有客户端都无任务，则结束
-                if all_client_is_waiting and self.spider.all_distribute_thread_is_done():
+                if all_client_is_waiting and self.master_distribute_thread_is_done():
                     logger.debug("所有客户端都已挂起，即将停止")
                     self.stop_all_client()
                     break
@@ -99,14 +107,22 @@ class ClientHeart:
         """
         from palp.conn import redis_conn
 
-        while not self.all_client_is_waiting:
+        while not self.stop and not self.all_client_is_waiting:
+            # 检测是否分发完毕
+            heart = {
+                "time": int(time.time()),
+                "waiting": self.spider.all_spider_controller_is_waiting(),
+            }
+            if self.spider.spider_master:
+                heart.update({'distribute_done': self.spider.all_distribute_thread_is_done()})
+            else:
+                heart.update({'distribute_done': False})
+
+            # 设置 redis 心跳
             redis_conn.hset(
                 settings.REDIS_KEY_HEARTBEAT,
                 self.client_name,
-                json.dumps({
-                    "time": int(time.time()),
-                    "waiting": self.spider.all_spider_controller_is_waiting(),
-                }, ensure_ascii=False)
+                json.dumps(heart, ensure_ascii=False)
             )
             time.sleep(self.beating_time)
 
@@ -123,6 +139,21 @@ class ClientHeart:
                 redis_conn.hdel(settings.REDIS_KEY_HEARTBEAT, self.client_name)
                 break
 
+    @staticmethod
+    def master_distribute_thread_is_done() -> bool:
+        """
+        判断 master 有没有把任务分发完毕
+
+        注意：master 死机后 slave 接管 master 会直接设置为 False
+        :return:
+        """
+        from palp.conn import redis_conn
+
+        master_detail = json.loads(redis_conn.get(settings.REDIS_KEY_MASTER).decode())
+        res = redis_conn.hget(settings.REDIS_KEY_HEARTBEAT, master_detail['name']).decode()
+
+        return json.loads(res)['distribute_done']
+
     @staticmethod
     def stop_all_client():
         """
@@ -176,11 +207,17 @@ class ClientHeart:
                     continue
                 else:
 
+                    # master 重新起名字
                     if self.spider.spider_master:
-                        redis_conn.set(settings.REDIS_KEY_MASTER, name)
+                        master_detail = json.loads(redis_conn.get(settings.REDIS_KEY_MASTER).decode())
+                        master_detail['name'] = name
+                        redis_conn.set(settings.REDIS_KEY_MASTER, json.dumps(master_detail, ensure_ascii=False))
 
+                    # 发送起始心跳
                     redis_conn.hset(settings.REDIS_KEY_HEARTBEAT, name, json.dumps({
                         "time": int(time.time()),
-                        "waiting": False,
+                        "waiting": False,  # 是否所有线程没事干
+                        'distribute_done': False,  # 任务分发是否结束
                     }, ensure_ascii=False))
+
                     return name
